#!/usr/bin/env python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

from __future__ import print_function

import imp, sys, os
import time
import logging
import cPickle as pkl
import gzip
import argparse

from adenine.core import analyze_results
from adenine.utils import extra

def main(dumpfile):
    # Load the configuration file
    config_path = os.path.dirname(dumpfile)
    config_path = os.path.join(os.path.abspath(config_path), 'ade_config.py')
    config = imp.load_source('ade_config', config_path)
    extra.set_module_defaults(config, {'file_format': 'pdf',
                                       'plotting_context': 'paper'})

    # Read the variables from the config file
    feat_names, class_names = config.feat_names, config.class_names
    # Load the results used with ade_run.py
    try:
        with gzip.open(os.path.join(os.path.dirname(dumpfile),'__data.pkl.tz'), 'r') as f:
            data = pkl.load(f)
            X, y = data['X'], data['y']
    except:
        sys.stderr("Cannot load __data.pkl.tz. Reloading data from config file ...")
        X, y = config.X, config.y

    # Initialize the log file
    fileName = 'results_'+os.path.basename(dumpfile)[0:-7]
    logFileName = os.path.join(os.path.dirname(dumpfile), fileName+'.log')
    logging.basicConfig(filename=logFileName, level=logging.INFO, filemode='w',
                        format='%(levelname)s (%(name)s): %(message)s')
    root_logger = logging.getLogger()
    ch = logging.StreamHandler()
    ch.setLevel(logging.ERROR)
    ch.setFormatter(logging.Formatter('%(levelname)s (%(name)s): %(message)s'))
    root_logger.addHandler(ch)

    tic = time.time()
    print("\nUnpickling output ...", end=' ')
    # Load the results
    with gzip.open(dumpfile, 'r') as f:
        res = pkl.load(f)

    print("done: {} s".format(extra.sec_to_time(time.time()-tic)))

    # Analyze the pipelines
    analyze_results.analyze(input_dict=res, root=os.path.dirname(dumpfile),
                            y=y, feat_names=feat_names, class_names=class_names,
                            plotting_context=config.plotting_context,
                            file_format=config.file_format)

# ----------------------------  RUN MAIN ---------------------------- #
if __name__ == '__main__':
    from adenine import __version__
    parser = argparse.ArgumentParser(#usage="%(prog)s RESULTS_DIR",
                                     description='Adenine script for analysing pipelines.')
    parser.add_argument('--version', action='version', version='%(prog)s v'+__version__)
    parser.add_argument("result_folder", help="specify results directory")
    args = parser.parse_args()
    root_folder = args.result_folder
    filename = [f for f in os.listdir(root_folder) if os.path.isfile(os.path.join(root_folder, f)) and f.endswith('.pkl.tz') and f !=  "__data.pkl.tz"]
    if not filename:
        sys.stderr.write("No .pkl file found in {}. Aborting...".format(root_folder))
        sys.exit(-1)

    # print("Starting the analysis of {}".format(filename))
    main(os.path.join(os.path.abspath(root_folder), filename[0])) # Run analysis



#!/usr/bin/python -W ignore::DeprecationWarning
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import os
import logging
import cPickle as pkl
import numpy as np
import pandas as pd
import matplotlib; matplotlib.use('AGG')
import seaborn as sns
import multiprocessing as mp

from sklearn import metrics

from adenine.core import plotting
from adenine.core.plotting import *
from adenine.utils.extra import title_from_filename
from adenine.utils.extra import timed, items_iterator

GLOBAL_INFO = ''  # to save info before logging is loaded


def est_clst_perf(root, data_in, labels=None, t_labels=None, model=(),
                  metric='euclidean'):
    """Estimate the clustering performance.

    This function estimate the clustering performance by means of several
    indexes. Then saves the results in a tree-like structure in the root folder.

    Parameters
    -----------
    root : string
        The root path for the output creation.

    data_in : array of float, shape : (n_samples, n_dimensions)
        The low space embedding estimated by the dimensinality reduction and
        manifold learning algorithm.

    labels : array of float, shape : n_samples
        The label assignment performed by the clustering algorithm.

    t_labels : array of float, shape : n_samples
        The true label vector; None if missing.

    model : sklearn or sklearn-like object
        An instance of the class that evaluates a step. In particular this must
        be a clustering model provided with the clusters_centers_ attribute
        (e.g. KMeans).

    metric : string
        The metric used during the clustering algorithms.
    """
    perf_out = dict()
    try:
        if hasattr(model, 'inertia_'):
            # Sum of distances of samples to their closest cluster center.
            perf_out['inertia'] = model.inertia_

        perf_out['silhouette'] = metrics.silhouette_score(data_in, labels, metric=metric)
        if t_labels is not None:
            # the next indexes need a gold standard
            perf_out['ari'] = metrics.adjusted_rand_score(t_labels, labels)
            perf_out['ami'] = metrics.adjusted_mutual_info_score(t_labels, labels)
            perf_out['homogeneity'] = metrics.homogeneity_score(t_labels, labels)
            perf_out['completeness'] = metrics.completeness_score(t_labels, labels)
            perf_out['v_measure'] = metrics.v_measure_score(t_labels, labels)

    except ValueError as e:
        logging.info("Clustering performance evaluation failed for {}. "
                     "Error: {}".format(model, e))
        # perf_out = {'empty': 0.0}
        perf_out['###'] = 0.

    # Define the filename
    filename = os.path.join(root, os.path.basename(root))
    with open(filename+'_scores.txt', 'w') as f:
        f.write("------------------------------------\n")
        f.write("Adenine: Clustering Performance for \n")
        f.write("\n")
        f.write(title_from_filename(root, " --> ") + "\n")
        f.write("------------------------------------\n")
        f.write("Index Name{}|{}Index Score\n".format(' '*10, ' '*4))
        f.write("------------------------------------\n")
        for elem in sorted(perf_out.keys()):
            f.write("{}{}|{}{:.4}\n"
                    .format(elem, ' '*(20-len(elem)), ' '*4, perf_out[elem]))
            f.write("------------------------------------\n")

    # pkl Dump
    with open(filename+'_scores.pkl', 'w+') as f:
        pkl.dump(perf_out, f)
    logging.info("Dumped : {}".format(filename+'_scores.pkl'))


def get_step_attributes(step, pos):
    """Get the attributes of the input step.

    This function returns the attributes (i.e. level, name, outcome) of the
    input step. This comes handy when dealing with steps with more than one
    parameter (e.g. KernelPCA 'poly' or 'rbf').

    Parameters
    -----------
    step : list
        A step coded by ade_run.py as
        [name, level, param, data_out, data_in, mdl obj, voronoi_mdl_obj]

    pos : int
        The position of the step inside the pipeline.

    Returns
    -----------
    name : string
        A unique name for the step (e.g. KernelPCA_rbf).

    level : {imputing, preproc, dimred, clustering}
        The step level.

    data_out : array of float, shape : (n_samples, n_out)
        Where n_out is n_dimensions for dimensionality reduction step, or 1
        for clustering.

    data_in : array of float, shape : (n_samples, n_in)
        Where n_in is n_dimensions for preprocessing/imputing/dimensionality
        reduction step, or n_dim for clustering (because the data have already
        been dimensionality reduced).

    param : dictionary
        The parameters of the sklearn object implementing the algorithm.

    mdl_obj : sklearn or sklearn-like object
        This is an instance of the class that evaluates a step.
    """
    name = step[0]
    level = step[1]  # {imputing, preproc, dimred, clustering}
    param = step[2]
    data_out = step[3]
    data_in = step[4]
    mdl_obj = step[5]
    voronoi_mdl_obj = step[6]
    if level.lower() == 'none':
        if pos == 0:
            level = 'preproc'
        elif pos == 1:
            level = 'dimred'
    metric = 'euclidean'

    # Imputing level
    if param.get('missing_values', ''):
        name += '-'+param['missing_values']
    if param.get('strategy', ''):
        name += '_'+param['strategy']

    # Preprocessing level
    if param.get('norm', ''):  # normalize
        name += '_'+param['norm']
    elif param.get('feature_range', ''):  # minmax
        name += "_({} - {})".format(*param['feature_range'])

    # Append additional parameters in the step name
    if name == 'KernelPCA':
        name += '_'+param['kernel']
    elif name == 'LLE':
        name += '_'+param['method']
    elif name == 'MDS':
        if param['metric']:
            name += '_metric'
        else:
            name += '_nonmetric'
    elif name == 'Hierarchical':
        name += '_'+param['affinity']+'_'+param['linkage']

    if param.get('affinity', '') == 'precomputed':
        metric = 'precomputed'

    # n_clusters = param.get('n_clusters', 0) or param.get('best_estimator_', dict()).get('n_clusters', 0) or param.get('best_estimator_', dict()).get('clusters_centers_', np.array([])).shape[0]
    n_clusters = param.get('n_clusters', 0) or  \
                 param.get('best_estimator_', dict()).get('cluster_centers_',
                           np.empty(0)).shape[0] or \
                 param.get('cluster_centers_', np.empty(0)).shape[0] or \
                 mdl_obj.__dict__.get('cluster_centers_', np.empty(0)).shape[0]
    if n_clusters > 0:
        name += '_' + str(n_clusters) + '-clusts'

    logging.info("{} : {}".format(level, name))
    return name, level, param, data_out, data_in, mdl_obj, voronoi_mdl_obj, metric


def make_df_clst_perf(root):
    """Summarize all the clustering performance estimations.

    Given the output file produced by est_clst_perf(), this function groups all
    of them together in friendly text and latex files, and saves the two files
    produced in a tree-like structure in the root folder.

    Parameters
    -----------
    root : string
        The root path for the output creation.
    """
    measures = ('ami', 'ari', 'completeness', 'homogeneity', 'v_measure',
                'inertia', 'silhouette')
    df = pd.DataFrame(columns=['pipeline']+list(measures))
    for root, directories, filenames in os.walk(root):
        for fn in filenames:
            if fn.endswith('_scores.pkl'):
                with open(os.path.join(root, fn), 'r') as f:
                    perf_out = pkl.load(f)
                perf_out['pipeline'] = title_from_filename(root, step_sep=" --> ")
                df = df.append(perf_out, ignore_index=True)
    df = df.fillna('')

    pipe_header = 'preprocess --> dim red --> clustering'
    size_pipe = max([len(p) for p in df['pipeline']]+[len(pipe_header)])
    size_ami, size_ari, size_com, size_hom, \
    size_vme, size_ine, size_sil = [2 +
        max([len('{: .3}'.format(p)) if p != '' else 3 for p in df[__]] + [len(__)])
            for __ in measures]

    # find the best value for each score
    best_scores = {__: max([p for p in df[__] if p != ''] or [np.nan]) for __ in measures}

    with open(os.path.join(root,'summary_scores.txt'), 'w') as f, \
         open(os.path.join(root,'summary_scores.tex'), 'w') as g:
        header = "{}{}|{}ami  |{}ari  |{}completeness  |{}homogeneity  |{}v_measure  |{}inertia  |{}silhouette  \n" \
            .format(pipe_header, ' '*(size_pipe-len(pipe_header)),
                    ' '*(size_ami-5), ' '*(size_ari-5),
                    ' '*(size_com-len("completeness  ")),
                    ' '*(size_hom-len("homogeneity  ")),
                    ' '*(size_vme-len("v_measure  ")),
                    ' '*(size_ine-len("inertia  ")),
                    ' '*(size_sil-len("silhouette  ")))
        f.write("-"*len(header) + "\n")
        f.write("Adenine: Clustering Performance for each pipeline\n")
        f.write("-"*len(header) + "\n")
        f.write(header)
        f.write("-"*len(header) + "\n")

        g.write(r"\documentclass{article}" "\n"
                r"\usepackage{adjustbox}" "\n"
                r"\usepackage{caption}" "\n"
                r"\captionsetup[table]{skip=10pt}" "\n"
                r"\begin{document}" "\n"
                r"\begin{table}[h!]" "\n"
                r"\centering" "\n"
                r"\caption{Adenine: Clustering Performance for each pipeline}" "\n"
                r"\label{clust-perf}" "\n"
                r"\begin{adjustbox}{max width=\textwidth}" "\n"
                r"\begin{tabular}{l|rc|rc|rc|rc|rc|rc|rc}" "\n"
                r"\textbf{preprocess $\to$ dim red $\to$ clustering} & \textbf{ami} "
                r"&& \textbf{ari} && \textbf{completeness} && \textbf{homogeneity} "
                r"&& \textbf{v\_measure} && \textbf{inertia} && \textbf{silhouette}"
                r" & \\ \hline" "\n")

        for _ in df.iterrows():
            row = _[1]
            ami, ari, com, hom, vme, ine, sil = ['{: .3}'.format(row[__])
                                if row[__] != '' else '---' for __ in measures]

            star = {__ : ' *' if row[__] == best_scores[__] else '  ' for __ in measures}
            f.write("{}{}|{}{}{}|{}{}{}|{}{}{}|{}{}{}|{}{}{}|{}{}{}|{}{}{}\n"
                 .format(row['pipeline'],' '*(size_pipe-len(row['pipeline'])),
                 ' '*(abs(size_ami-len(str(ami))-2)), ami, star['ami'],
                 ' '*(abs(size_ari-len(str(ari))-2)), ari, star['ari'],
                 ' '*(abs(size_com-len(str(com))-2)), com, star['completeness'],
                 ' '*(abs(size_hom-len(str(hom))-2)), hom, star['homogeneity'],
                 ' '*(abs(size_vme-len(str(vme))-2)), vme, star['v_measure'],
                 ' '*(abs(size_ine-len(str(ine))-2)), ine, star['inertia'],
                 ' '*(abs(size_sil-len(str(sil))-2)), sil, star['silhouette']
            ))

            g.write(r"{} & {}&{} & {}&{} & {}&{} & {}&{} & {}&{} & {}&{} & {}&{} \\" "\n"
                .format(
                row['pipeline'].replace('-->', r'$\to$'),
                ami, star['ami'],
                ari, star['ari'],
                com, star['completeness'],
                hom, star['homogeneity'],
                vme, star['v_measure'],
                ine, star['inertia'],
                sil, star['silhouette'],
            ))

        f.write("-"*len(header) + "\n")
        g.write(r"\hline" "\n"
                r"\end{tabular}" "\n"
                r"\end{adjustbox}" "\n"
                r"\end{table}" "\n"
                r"\end{document}")
    # df.to_csv(os.path.join(rootFolder,'all_scores.csv'), na_rep='-', index_label=False, index=False)


def analysis_worker(elem, root, y, feat_names, class_names, lock):
    """Parallel pipelines analysis.

    Parameters
    ----------
    elem : list
        The first two element of this list are the pipe_id and all the data of
        that pipeline.

    root : string
        The root path for the output creation.

    y : array of float, shape : n_samples
        The label vector; None if missing.

    feat_names : array of integers (or strings), shape : n_features
        The feature names; a range of numbers if missing.

    class_names : array of integers (or strings), shape : n_features
        The class names; a range of numbers if missing.

    lock : multiprocessing.synchronize.Lock
        Obtained by multiprocessing.Lock().
        Needed for optional creation of directories.
    """
    # Getting pipeID and content
    pipe, content = elem[:2]

    out_folder = ''  # where the results will be placed
    logging.info("------\n{} : \n".format(pipe))
    for i, step in enumerate(sorted(content.keys())):
        # Tree-like folder structure definition
        step_name, step_level, step_param, step_out, step_in, mdl_obj, \
            voronoi_mdl_obj, metric = get_step_attributes(content[step], pos=i)

        # Output folder definition & creation
        out_folder = os.path.join(out_folder, step_name)
        rootname = os.path.join(root, out_folder)
        with lock:
            if not os.path.exists(rootname):
                os.makedirs(rootname)

        # Launch analysis
        if step_level == 'dimred':
            make_scatter(root=rootname, data_in=step_out, labels=y, true_labels=True)
            make_silhouette(root=rootname, labels=y, data_in=step_out, model=mdl_obj)

            if hasattr(mdl_obj, 'explained_variance_ratio_'):
                plot_PCmagnitude(root=rootname,
                                 points=mdl_obj.explained_variance_ratio_,
                                 title='Explained variance ratio')
            if hasattr(mdl_obj, 'lambdas_'):
                plot_PCmagnitude(root=rootname,
                                 points=mdl_obj.lambdas_/np.sum(mdl_obj.lambdas_),
                                 title='Normalized eigenvalues of the centered kernel matrix')
        if step_level == 'clustering':
            if hasattr(mdl_obj, 'affinity_matrix_'):
                n_clusters = mdl_obj.__dict__.get('cluster_centers_', np.empty(0)).shape[0]
                if hasattr(mdl_obj, 'n_clusters'):
                    n_clusters = mdl_obj.n_clusters

                plot_eigs(root=rootname, affinity=mdl_obj.affinity_matrix_,
                          n_clusters=n_clusters,
                          title='Eigenvalues of the graph associated to the affinity matrix')
            if hasattr(mdl_obj, 'cluster_centers_'):
                _est_name = mdl_obj.__dict__.get('estimator_name', '') or type(mdl_obj).__name__
                if _est_name != 'AffinityPropagation':
                    # disable the voronoi plot for affinity prop
                    make_voronoi(root=rootname, labels=y, data_in=step_in,
                                 model=voronoi_mdl_obj)
            elif hasattr(mdl_obj, 'n_leaves_'):
                make_tree(root=rootname, data_in=step_in, labels=step_out, model=mdl_obj)
                make_dendrogram(root=rootname, data_in=step_in, labels=y, model=mdl_obj)

            make_scatter(root=rootname, labels=step_out, data_in=step_in, model=mdl_obj)
            make_silhouette(root=rootname, labels=step_out, data_in=step_in, model=mdl_obj)
            est_clst_perf(root=rootname, data_in=step_in, labels=step_out,
                          t_labels=y, model=mdl_obj, metric=metric)


@timed
def analyze(input_dict, root, y=None, feat_names=(), class_names=(), **kwargs):
    """Analyze the results of ade_run.

    This function analyze the dictionary generated by ade_run, generates the
    plots, and saves them in a tree-like folder structure in rootFolder.

    Parameters
    -----------
    input_dict : dictionary
        The dictionary created by ade_run.py on some data.

    root : string
        The root path for output creation.

    y : array of float, shape : n_samples
        The label vector; None if missing.

    feat_names : array of integers (or strings), shape : n_features
        The feature names; a range of numbers if missing.

    class_names : array of integers (or strings), shape : n_features
        The class names; a range of numbers if missing.

    kwargs : dictionary
        Additional optional parameters. In particular it can contain
        'plotting_context' and 'file_format' variables, if specified in
        the config file.
    """
    if GLOBAL_INFO:
        logging.info(GLOBAL_INFO)
    if kwargs.get('plotting_context', None):
        sns.set_context(kwargs.get('plotting_context'))

    FILE_FORMATS = ('png', 'pdf')
    ff = kwargs.get('file_format', FILE_FORMATS[0]).lower()

    if ff not in FILE_FORMATS:
        logging.warning("File format unknown. "
                        "Please select one of {}".format(FILE_FORMATS))
        plotting.set_file_ext(FILE_FORMATS[0])
    else:
        plotting.set_file_ext(ff)
    logging.info("File format set to {}".format(plotting.GLOBAL_FF))
    lock = mp.Lock()
    # Parallel(n_jobs=len(inputDict))(delayed(analysis_worker)(elem,rootFolder,y,feat_names,class_names,lock) for elem in inputDict.iteritems())
    ps = []
    for elem in items_iterator(input_dict):
        p = mp.Process(target=analysis_worker,
                       args=(elem, root, y, feat_names, class_names, lock))
        p.start()
        ps.append(p)

    for p in ps:
        p.join()

    make_df_clst_perf(root)



#!/usr/bin/python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

def new_fun(arg1 = 'Default', arg2 = 'Default'):
    """Short explanation.

    This is the very long explanation

    Parameters
    -----------
    arg1 : type, default : 'Default'
        What is arg1.

    arg2 : {'Default', 'Different', 'Another'}
        What is arg2.

    Returns
    -----------
    out : type
        What is out.
    """



#!/usr/bin/python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import os
import time
import seaborn as sns

from datetime import datetime
from itertools import product

palette = sns.color_palette("Set1")

def get_color(i=0):
    return palette[i]

def next_color():
    palette.append(palette.pop(0))
    return palette[-1]

def reset_palette(n_colors=6):
    global palette
    palette = sns.color_palette("Set1", n_colors)

# ensure_list = lambda x: x if type(x) == list else [x]
def ensure_list(x):
    return x if type(x) == list else [x]

def values_iterator(dictionary):
    '''Add support for python2 or 3 dictionary iterators. '''
    try:
        v = dictionary.itervalues() # python 2
    except:
        v = dictionary.values() # python 3
    return v

def items_iterator(dictionary):
    '''Add support for python2 or 3 dictionary iterators. '''
    try:
        gen = dictionary.iteritems() # python 2
    except:
        gen = dictionary.items() # python 3
    return gen

def modified_cartesian(*args, **kwargs):
    """Modified Cartesian product.

    This function takes two (ore more) lists and returns their Cartesian product,
    if one of the two list is empty this function returns the non-empty one.

    Parameters
    -----------
    *args : lists, length : two or more
        The group of input lists.

    Returns
    -----------
    cp : list
        The Cartesian Product of the two (or more) nonempty input lists.
    """
    # Get the non-empty input lists
    if kwargs.get('pipes_mode', False):
        nonempty = [ensure_list(arg) for arg in args if len(ensure_list(arg)) > 0]
    else:
        nonempty = [ensure_list(arg) if len(ensure_list(arg)) > 0 else [None] for arg in args]

    # Cartesian product
    return [list(c) for c in product(*nonempty)]

def make_time_flag():
    """Generate a time flag.

    This function simply generates a time flag using the current time.

    Returns
    -----------
    timeFlag : string
        A unique time flag.
    """
    y = str(time.localtime().tm_year)
    mo = str(time.localtime().tm_mon)
    d = str(time.localtime().tm_mday)
    h = str(time.localtime().tm_hour)
    mi = str(time.localtime().tm_min)
    s = str(time.localtime().tm_sec)
    return h+':'+mi+':'+s+'_'+d+'-'+mo+'-'+y

def sec_to_time(seconds):
    """Transform seconds into a formatted time string.

    Parameters
    -----------
    seconds : int
        Seconds to be transformed.

    Returns
    -----------
    time : string
        A well formatted time string.
    """
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    return "%02d:%02d:%02d" % (h, m, s)

def get_time():
    return datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d_%H.%M.%S')

def title_from_filename(root, step_sep="$\mapsto$"):
    # Define the plot title. List is smth like ['results', 'ade_debug_', 'Standardize', 'PCA']
    i = [i for i, s in enumerate(root.split(os.sep)) if 'ade_' in s][0]

    # lambda function below does: ('a_b_c') -> 'c b a'
    return step_sep.join(map(lambda x: ' '.join(x.split('_')[::-1]), root.split(os.sep)[i+1:]))

def ensure_symmetry(X):
    """Ensure matrix symmetry.

    Parameters
    -----------
    X : numpy.ndarray
        Input matrix of precomputed pairwise distances.

    Returns
    -----------
    new_X : numpy.ndarray
        Symmetric distance matrix. Values are averaged.
    """
    if not (X.T == X).all():
        return (X.T + X) / 2.

def timed(function):
    """Decorator that measures wall time of the decored function."""
    def timed_function(*args, **kwargs):
        t0 = time.time()
        result = function(*args, **kwargs)
        print("\nAdenine {} - Elapsed time : {} s\n".format(function.__name__, sec_to_time(time.time() - t0)))
        return result
    return timed_function

def set_module_defaults(module, dictionary):
    """Set default variables of a module, given a dictionary.
    Used after the loading of the configuration file to set some defaults."""
    for k, v in items_iterator(dictionary):
        try:
            getattr(module, k)
        except AttributeError:
            setattr(module, k, v)



#!/usr/bin/python -W ignore::DeprecationWarning
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import os
import logging
# import cPickle as pkl
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns; sns.set(font="monospace")
from sklearn import metrics
# Legacy import
try:
    from sklearn.model_selection import StratifiedShuffleSplit
except ImportError:
    from sklearn.cross_validation import StratifiedShuffleSplit

from adenine.utils.extra import \
    (next_color, reset_palette, title_from_filename, items_iterator)

__all__ = ["make_silhouette", "make_scatter", "make_voronoi", "make_tree",
           "make_dendrogram", "plot_PCmagnitude", "plot_eigs"]

GLOBAL_FF = 'png'


def set_file_ext(ext):
    global GLOBAL_FF
    GLOBAL_FF = ext


def make_silhouette(root, data_in, labels, model=()):
    """Generate and save the silhouette plot of data_in w.r.t labels.

    This function generates the silhouette plot representing how data are
    correctly clustered, based on labels.
    The plots will be saved into the root folder in a tree-like structure.

    Parameters
    -----------
    root : string
        The root path for the output creation

    data_in : array of float, shape : (n_samples, n_dimensions)
        The low space embedding estimated by the dimensionality reduction and
        manifold learning algorithm.

    labels : array of float, shape : n_samples
        The label vector. It can contain true or estimated labels.

    model : sklearn or sklearn-like object
        An instance of the class that evaluates a step.
    """
    # Create a subplot with 1 row and 2 columns
    fig, (ax1) = plt.subplots(1, 1)
    fig.set_size_inches(20, 15)

    # The silhouette coefficient can range from -1, 1
    # ax1.set_xlim([-1, 1])

    # The (n_clusters+1)*10 is for inserting blank space between silhouette
    # plots of individual clusters.
    n_clusters = np.unique(labels).shape[0]
    ax1.set_ylim([0, len(data_in) + (n_clusters + 1) * 10])

    # The silhouette_score gives the average value for all the samples.
    # This gives a perspective into the density and separation of the formed
    # clusters
    metric = model.affinity if hasattr(model, 'affinity') else 'euclidean'
    sample_silhouette_values = metrics.silhouette_samples(data_in, labels, metric=metric)
    sil = np.mean(sample_silhouette_values)

    y_lower = 10
    reset_palette()
    for i, label in enumerate(np.unique(labels)):
        # Aggregate the silhouette scores for samples belonging to
        # cluster i, and sort them
        ith_cluster_silhouette_values = sample_silhouette_values[labels == label]
        ith_cluster_silhouette_values.sort()

        size_cluster_i = ith_cluster_silhouette_values.shape[0]
        y_upper = y_lower + size_cluster_i

        # color = cm.spectral(float(i) / n_clusters)
        color = next_color()
        ax1.fill_betweenx(np.arange(y_lower, y_upper),
                          0, ith_cluster_silhouette_values,
                          facecolor=color, edgecolor=color, alpha=0.7)

        # Label the silhouette plots with their cluster numbers at the middle
        ax1.text(-0.05, y_lower + 0.5 * size_cluster_i, str(label))

        # Compute the new y_lower for next plot
        y_lower = y_upper + 10  # 10 for the 0 samples

    # ax1.set_title("Silhouette plot for the various clusters.")
    ax1.set_xlabel("silhouette coefficient values")
    ax1.set_ylabel("cluster label")

    # The vertical line for average silhoutte score of all the values
    ax1.axvline(x=sil, color="red", linestyle="--")
    ax1.set_yticks([])
    # ax1.set_xticks([-0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6, 0.8, 1])

    plt.suptitle("Silhouette analysis. "
                 "{0} clusters for {2} samples, average score {1:.4f}"
                 .format(n_clusters, sil, data_in.shape[0]))

    filename = os.path.join(root, os.path.basename(root)+"_silhouette."+GLOBAL_FF)
    fig.savefig(filename)
    logging.info('Figured saved {}'.format(filename))
    plt.close()


def make_scatter(root, data_in, labels=None, true_labels=False, model=()):
    """Generates and saves the scatter plot of the dimensionality reduced data set.

    This function generates the scatter plot representing the dimensionality
    reduced data set. The plots will be saved into the root folder in a
    tree-like structure.

    Parameters
    ----------
    root : string
        The root path for the output creation

    data_in : array of float, shape : (n_samples, n_dimensions)
        The low space embedding estimated by the dimensinality reduction and
        manifold learning algorithm.

    labels : array of float, shape : n_samples
        The label vector. It can contain true or estimated labels.

    true_labels : boolean
        Identify if labels contains true or estimated labels.

    model : sklearn or sklearn-like object
        An instance of the class that evaluates a step. In particular this must
        be a clustering model provided with the clusters_centers_ attribute
        (e.g. KMeans).
    """
    n_samples, n_dim = data_in.shape

    # Define plot color
    if labels is None:
        y = np.zeros((n_samples))
        _hue = ' '
    else:
        y = labels
        _hue = 'Classes' if true_labels else 'Estimated Labels'

    title = title_from_filename(root)

    # Seaborn scatter plot
    # 2D plot
    X = data_in[:, :2]
    idx = np.argsort(y)

    # df = pd.DataFrame(data=np.hstack((X[idx,:2],y[idx,np.newaxis])), columns=["$x_1$","$x_2$",_hue])
    df = pd.DataFrame(data=np.hstack((X[idx, :2], y[idx][:, np.newaxis])),
                      columns=["$x_1$", "$x_2$", _hue])
    if df.dtypes[_hue] != 'O': df[_hue] = df[_hue].astype('int64')
    # Generate seaborn plot
    g = sns.FacetGrid(df, hue=_hue, palette="Set1", size=5, legend_out=False)
    g.map(plt.scatter, "$x_1$", "$x_2$", s=100, linewidth=.5, edgecolor="white")
    if _hue != ' ': g.add_legend() #!! customize legend
    # g.set_xticklabels([])
    # g.set_yticklabels([])
    g.ax.autoscale_view(True, True, True)
    plt.title(title)
    filename = os.path.join(root, os.path.basename(root)+"_scatter2D."+GLOBAL_FF)
    plt.savefig(filename)
    logging.info('Figured saved {}'.format(filename))
    plt.close()

    # 3D plot
    filename = os.path.join(root,os.path.basename(root)+"_scatter3D."+GLOBAL_FF)
    X = data_in[:, :3]
    if X.shape[1] < 3:
        logging.warning('{} not generated (data have less than 3 dimensions)'
                        .format(filename))
    else:
        try:
            from mpl_toolkits.mplot3d import Axes3D
            ax = plt.figure().gca(projection='3d')
            # ax.scatter(X[:,0], X[:,1], X[:,2], y, c=y, cmap='hot', s=100, linewidth=.5, edgecolor="white")
            y = np.array(y)
            reset_palette(len(np.unique(y)))
            for _, label in enumerate(np.unique(y)):
                idx = np.where(y == label)[0]
                ax.plot(X[:, 0][idx], X[:, 1][idx], X[:, 2][idx], 'o',
                        c=next_color(), label=str(label), mew=.5, mec="white")

            ax.set_xlabel(r'$x_1$')
            ax.set_ylabel(r'$x_2$')
            ax.set_zlabel(r'$x_3$')
            ax.autoscale_view(True, True, True)
            ax.set_title(title)
            ax.legend(loc='upper left', numpoints=1, ncol=10, fontsize=8,
                      bbox_to_anchor=(0, 0))
            # plt.legend(loc='upper left', numpoints=1, ncol=3, fontsize=8, bbox_to_anchor=(0, 0111))
            plt.savefig(filename)
            logging.info('Figured saved {}'.format(filename))
            plt.close()
        except Exception as e:
            logging.info('Error in 3D plot: ' + str(e))

    # seaborn pairplot
    n_cols = min(data_in.shape[1], 3)
    cols = ["$x_{}$".format(i+1) for i in range(n_cols)]
    X = data_in[:, :3]
    idx = np.argsort(y)
    df = pd.DataFrame(data=np.hstack((X[idx,:],y[idx,np.newaxis])),
                      columns=cols+[_hue])
    if df.dtypes[_hue] != 'O': df[_hue] = df[_hue].astype('int64')
    g = sns.PairGrid(df, hue=_hue, palette="Set1", vars=cols)
    g = g.map_diag(plt.hist)  # , palette="Set1")
    g = g.map_offdiag(plt.scatter, s=100, linewidth=.5, edgecolor="white")

    # g = sns.pairplot(df, hue=_hue, palette="Set1", vars=["$x_1$","$x_2$","$x_3$"]), size=5)
    if _hue != ' ': plt.legend(title=_hue,bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0., fontsize="large")
    plt.suptitle(title, x=0.6, y=1.01, fontsize="large")
    filename = os.path.join(root, os.path.basename(root)+"_pairgrid."+GLOBAL_FF)
    g.savefig(filename)
    logging.info('Figured saved {}'.format(filename))
    plt.close()


def make_voronoi(root, data_in, labels=None, true_labels=False, model=()):
    """Generate and save the Voronoi tessellation obtained from the clustering
    algorithm.

    This function generates the Voronoi tessellation obtained from the
    clustering algorithm applied on the data projected on a two-dimensional
    embedding. The plots will be saved into the appropriate folder of the
    tree-like structure created into the root folder.

    Parameters
    -----------
    root : string
        The root path for the output creation

    data_in : array of float, shape : (n_samples, n_dimensions)
        The low space embedding estimated by the dimensinality reduction and
        manifold learning algorithm.

    labels : array of int, shape : n_samples
        The result of the clustering step.

    true_labels : boolean [deprecated]
        Identify if labels contains true or estimated labels.

    model : sklearn or sklearn-like object
        An instance of the class that evaluates a step. In particular this must
        be a clustering model provided with the clusters_centers_ attribute
        (e.g. KMeans).
    """
    n_samples, n_dim = data_in.shape

    # Define plot color
    if labels is None:
        y = np.zeros((n_samples))
        _hue = ' '
    else:
        y = labels  # use the labels if provided
        _hue = 'Classes'

    title = title_from_filename(root)

    # Seaborn scatter Plot
    X = data_in[:, :2]
    idx = np.argsort(y)
    X = X[idx,:]
    y = y[idx, np.newaxis]
    df = pd.DataFrame(data=np.hstack((X, y)), columns=["$x_1$", "$x_2$", _hue])
    if df.dtypes[_hue] != 'O': df[_hue] = df[_hue].astype('int64')
    # Generate seaborn plot
    g = sns.FacetGrid(df, hue=_hue, palette="Set1", size=5, legend_out=False)
    g.map(plt.scatter, "$x_1$", "$x_2$", s=100, linewidth=.5, edgecolor="white")
    if _hue != ' ': g.add_legend() #!! customize legend
    g.ax.autoscale_view(True, True, True)
    plt.title(title)

    # Add centroids
    if hasattr(model, 'cluster_centers_'):
        plt.scatter(model.cluster_centers_[:, 0], model.cluster_centers_[:, 1],
                    s=100, marker='h', c='w')

    # Make and add to the Plot the decision boundary.
    npoints = 1000  # the number of points in that makes the background. Reducing this will decrease the quality of the voronoi background
    x_min, x_max = X[:, 0].min(), X[:, 0].max()
    y_min, y_max = X[:, 1].min(), X[:, 1].max()
    offset = (x_max - x_min) / 5. + (y_max - y_min) / 5.  # zoom out the plot
    xx, yy = np.meshgrid(np.linspace(x_min-offset, x_max+offset, npoints),
                         np.linspace(y_min-offset, y_max+offset, npoints))

    # Obtain labels for each point in mesh. Use last trained model.

    Z = model.predict(np.c_[xx.ravel(), yy.ravel()])
    # Put the result into a color plot
    Z = Z.reshape(xx.shape)

    plt.imshow(Z, interpolation='nearest',
               extent=(xx.min(), xx.max(), yy.min(), yy.max()),
               cmap=plt.get_cmap('Pastel1'), aspect='auto', origin='lower')

    plt.xlim([xx.min(), xx.max()])
    plt.ylim([yy.min(), yy.max()])

    filename = os.path.join(root, os.path.basename(root)+"."+GLOBAL_FF)
    plt.savefig(filename)
    logging.info('Figured saved {}'.format(filename))
    plt.close()


def make_tree(root, data_in, labels=None, model=()):
    """Generate and save the tree structure obtained from the clustering
    algorithm.

    This function generates the tree obtained from the clustering algorithm
    applied on the data. The plots will be saved into the appropriate folder of
    the tree-like structure created into the root folder.

    Parameters
    -----------
    root : string
        The root path for the output creation

    data_in : array of float, shape : (n_samples, n_dimensions)
        The low space embedding estimated by the dimensinality reduction and
        manifold learning algorithm.

    labels : array of int, shape : n_samples
        The result of the clustering step.

    model : sklearn or sklearn-like object
        An instance of the class that evaluates a step. In particular this must
        be a clustering model provided with the clusters_centers_ attribute
        (e.g. KMeans).
    """
    filename = os.path.join(root, os.path.basename(root)+'_tree.pdf')
    try:
        import itertools
        import pydot
        graph = pydot.Dot(graph_type='graph')

        reset_palette(len(np.unique(labels)))
        global palette
        colors = {v: k for k, v in items_iterator(dict(enumerate(np.unique(labels))))}
        ii = itertools.count(data_in.shape[0])
        for k, x in enumerate(model.children_):
            root_node = next(ii)
            left_node = pydot.Node(str(x[0]), style="filled",
                                   fillcolor=palette.as_hex()[colors[labels[x[0]]]] if x[0] < labels.shape[0] else 'white')
            right_node = pydot.Node(str(x[1]), style="filled",
                                    fillcolor=palette.as_hex()[colors[labels[x[1]]]] if x[1] < labels.shape[0] else 'white')

            graph.add_node(left_node)
            graph.add_node(right_node)
            graph.add_edge(pydot.Edge(root_node, left_node))
            graph.add_edge(pydot.Edge(root_node, right_node))

        # graph.write_png(filename[:-2]+"ng")
        graph.write_pdf(filename)
        logging.info('Figured saved {}'.format(filename))
    except Exception as e:
        logging.critical('Cannot create {}. tb: {}'.format(filename, e))


def make_dendrogram(root, data_in, labels, model=(), n_max=150):
    """Generate and save the dendrogram obtained from the clustering algorithm.

    This function generates the dendrogram obtained from the clustering
    algorithm applied on the data. The plots will be saved into the appropriate
    folder of the tree-like structure created into the root folder. The row
    colors of the heatmap are the either true or estimated data labels.

    Parameters
    -----------
    root : string
        The root path for the output creation

    data_in : array of float, shape : (n_samples, n_dimensions)
        The low space embedding estimated by the dimensinality reduction and
        manifold learning algorithm.

    labels : array of int, shape : n_samples
        The result of the clustering step.

    model : sklearn or sklearn-like object
        An instance of the class that evaluates a step. In particular this must
        be a clustering model provided with the clusters_centers_ attribute
        (e.g. KMeans).

    n_max : int, (INACTIVE)
        The maximum number of samples to include in the dendrogram.
        When the number of samples is bigger than n_max, only n_max samples
        randomly extracted from the dataset are represented. The random
        extraction is performed using
        sklearn.model_selection.StratifiedShuffleSplit
        (or sklearn.cross_validation.StratifiedShuffleSplit for legacy
        reasons).
    """
    # # Check for the number of samples
    # n_samples = data_in.shape[0]
    # if n_samples > n_max:
    #     try: # Legacy for sklearn
    #         sss = StratifiedShuffleSplit(_y, test_size=n_max, n_iter=1)
    #     except TypeError:
    #         sss = StratifiedShuffleSplit(n_iter=1, test_size=n_max).split(data_in, _y)
    #
    #     _, idx = list(sss)[0]
    #     data_in = data_in[idx, :]
    #     labels = labels[idx]
    #     trueLabel = trueLabel[idx]

    # tmp = np.hstack((np.arange(0,data_in.shape[0],1)[:,np.newaxis], data_in[:,0][:,np.newaxis], data_in[:,1][:,np.newaxis]))
    col = ["$x_{"+str(i)+"}$" for i in np.arange(0, data_in.shape[1], 1)]
    df = pd.DataFrame(data=data_in, columns=col)

    # -- Code for row colors adapted from:
    # https://stanford.edu/~mwaskom/software/seaborn/examples/structured_heatmap.html
    # Create a custom palette to identify the classes
    n_colors = len(set(labels))
    custom_pal = sns.color_palette("hls", n_colors)
    custom_lut = dict(zip(map(str, range(n_colors)), custom_pal))

    # Convert the palette to vectors that will be drawn on the side of the matrix
    custom_colors = pd.Series(map(str, labels)).map(custom_lut)

    # Create a custom colormap for the heatmap values
    cmap = sns.diverging_palette(200, 10, sep=1, n=15, center="dark", as_cmap=True)#[::-1]
    # ---------------------------------------- #

    if model.affinity == 'precomputed': # TODO fix me: fede
        # tmp is the distance matrix
        make_dendrograms = False
        if make_dendrograms:
            sns.set(font="monospace")
            for method in ['single','complete','average','weighted','centroid','median','ward']:
                from scipy.cluster.hierarchy import linkage
                # print("Compute linkage matrix with metric={} ...".format(method))
                Z = linkage(data_in, method=method, metric='euclidean')
                g = sns.clustermap(df.corr(), method=method, row_linkage=Z, col_linkage=Z)
                filename = os.path.join(root, '_'.join((os.path.basename(root), method, '_dendrogram.png')))
                g.savefig(filename)
                logging.info('Figured saved {}'.format(filename))
                plt.close()
        # avg_sil = True
        # if avg_sil:
        #     try:
        #         from ignet.plotting.silhouette_hierarchical import plot_avg_silhouette
        #         filename = plot_avg_silhouette(data_in)
        #         logging.info('Figured saved {}'.format(filename))
        #     except:
        #         logging.critical("Cannot import name ignet.plotting'))
        # return

    # workaround to a different name used for manhatta / cityblock distance
    if model.affinity == 'manhattan': model.affinity = 'cityblock'
    g = sns.clustermap(df, method=model.linkage, metric=model.affinity,
                       row_colors=custom_colors, linewidths=.5, cmap=cmap)

    plt.setp(g.ax_heatmap.yaxis.get_majorticklabels(), rotation=0, fontsize=5)
    filename = os.path.join(root, os.path.basename(root)+'_dendrogram.'+GLOBAL_FF)
    g.savefig(filename)
    logging.info('Figured saved {}'.format(filename))
    plt.close()


def plot_PCmagnitude(root, points, title='', ylabel=''):
    """Generate and save the plot representing the trend of principal
    components magnitude.

    Parameters
    -----------
    root : string
        The root path for the output creation.

    points : array of float, shape : n_components
        This could be the explained variance ratio or the eigenvalues of the
        centered matrix, according to the PCA algorithm of choice, respectively
        PCA or KernelPCA.

    title : string
        Plot title.

    ylabel : string
        Y-axis label.
    """
    plt.plot(np.arange(1, len(points)+1), points, '-o')
    plt.title(title)
    plt.grid('on')
    plt.ylabel(ylabel)
    plt.xlim([1, min(20, len(points)+1)])  # Show maximum 20 components
    plt.ylim([0, 1])
    filename = os.path.join(root, os.path.basename(root)+"_magnitude."+GLOBAL_FF)
    plt.savefig(filename)
    plt.close()


def plot_eigs(root, affinity, n_clusters=0, title='', ylabel='',
              normalised=True):
    """Generate and save the plot representing the eigenvalues of the Laplacian
    associated to data affinity matrix.

    Parameters
    -----------
    root : string
        The root path for the output creation.

    affinity : array of float, shape : (n_samples, n_samples)
        The affinity matrix.

    n_clusters : int, optional
        The number of clusters.

    title : string, optional
        Plot title.

    ylabel : string, optional
        Y-axis label.

    normalised : boolean, optional
        Choose whether to normalise the Laplacian matrix.
    """
    W = affinity - np.diag(np.diag(affinity))
    D = np.diag([np.sum(x) for x in W])
    L = D - W

    if normalised:
        # aux = np.linalg.inv(np.diag([np.sqrt(np.sum(x)) for x in W]))
        aux = np.diag(1. / np.array([np.sqrt(np.sum(x)) for x in W]))
        L = np.eye(L.shape[0]) - (np.dot(np.dot(aux, W), aux))  # normalised L

    try:
        w, v = np.linalg.eig(L)
        w = np.array(sorted(np.abs(w)))
        plt.plot(np.arange(1, len(w)+1), w, '-o')
        plt.title(title)
        plt.grid('on')
        plt.ylabel(ylabel)
        plt.xlim([1, min(20, len(w)+1)])  # Show maximum 20 components
        if n_clusters > 0:
            plt.axvline(x=n_clusters+.5, linestyle='--', color='r',
                        label='selected clusters')
        plt.legend(loc='upper right', numpoints=1, ncol=10, fontsize=8)
        # , bbox_to_anchor=(1, 1))
        filename = os.path.join(root, os.path.basename(root)+"_eigenvals."+GLOBAL_FF)
        plt.savefig(filename)
    except np.linalg.LinAlgError:
        logging.critical("Error in plot_eigs: Affinity matrix contained"
                         "negative values. You can try by specifying"
                         "normalised=False")
    plt.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import logging
# import numpy as np
from adenine.utils.extra import modified_cartesian, ensure_list, values_iterator

# from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import Normalizer
from sklearn.decomposition import PCA
from sklearn.decomposition import RandomizedPCA
from sklearn.decomposition import IncrementalPCA
from sklearn.decomposition import KernelPCA
from sklearn.manifold import Isomap
from sklearn.manifold import LocallyLinearEmbedding
from sklearn.manifold import SpectralEmbedding
from sklearn.manifold import MDS
from sklearn.manifold import TSNE
from sklearn.cluster import KMeans
from sklearn.cluster import AffinityPropagation
from sklearn.cluster import MeanShift
from sklearn.cluster import SpectralClustering
from sklearn.cluster import AgglomerativeClustering

from adenine.utils.extensions import DummyNone
from adenine.utils.extensions import Imputer
from adenine.utils.extensions import GridSearchCV
from adenine.utils.extensions import silhouette_score


def parse_preproc(key, content):
    """Parse the options of the preprocessing step.

    This function parses the preprocessing step coded as dictionary in the
    ade_config file.

    Parameters
    -----------
    key : {'None', 'Recenter', 'Standardize', 'Normalize', 'MinMax'}
        The type of selected preprocessing step.

    content : dict
        A dictionary containing parameters for each preprocessing
        class. Each parameter can be a list; in this case for each combination
        of parameters a different pipeline will be created.

    Returns
    -----------
    pptpl : tuple
        A tuple made like that ('preproc_name', preproc_obj), where preproc_obj
        is an sklearn 'transforms' (i.e. it has bot a .fit and .transform
        method).
    """
    if key.lower() == 'none':
        pp = DummyNone()
    elif key.lower() == 'recenter':
        pp = StandardScaler(with_mean=True, with_std=False)
    elif key.lower() == 'standardize':
        pp = StandardScaler(with_mean=True, with_std=True)
    elif key.lower() == 'normalize':
        content.setdefault('norm', 'l2')
        # pp = Normalizer(norm=content[1][0])
        pp = Normalizer(**content)
    elif key.lower() == 'minmax':
        content.setdefault('feature_range', (0, 1))
        pp = MinMaxScaler(**content)
    else:
        pp = DummyNone()
    return (key, pp)


def parse_dimred(key, content):
    """Parse the options of the dimensionality reduction step.

    This function does the same as parse_preproc but works on the
    dimensionality reduction & manifold learning options.

    Parameters
    -----------
    key : {'None', 'PCA', 'KernelPCA', 'Isomap', 'LLE', 'SE', 'MDS', 'tSNE'}
        The selected dimensionality reduction algorithm.

    content : dict
        A dictionary containing parameters for each dimensionality reduction
        class. Each parameter can be a list; in this case for each combination
        of parameters a different pipeline will be created.

    Returns
    -----------
    drtpl : tuple
        A tuple made like that ('dimres_name', dimred_obj), where dimred_obj is
        a sklearn 'transforms' (i.e. it has bot a .fit and .transform method).
    """
    drs = {'none': DummyNone, 'pca': PCA, 'incrementalpca': IncrementalPCA,
           'randomizedpca': RandomizedPCA, 'kernelpca': KernelPCA,
           'isomap': Isomap, 'lle': LocallyLinearEmbedding,
           'se': SpectralEmbedding, 'mds': MDS, 'tsne': TSNE}

    content.setdefault('n_components', 3)  # use three cluster as default
    dr = drs.get(key.lower(), DummyNone)(**content)
    return (key, dr)

    # if key.lower() == 'none':
    #     dr = DummyNone(**content)
    # elif key.lower() == 'pca':
    #     dr = PCA(**content) # this by default takes all the components it can
    # elif key.lower() == 'incrementalpca':
    #     dr = IncrementalPCA(**content)
    # elif key.lower() == 'randomizedpca':
    #     dr = RandomizedPCA(**content)
    # elif key.lower() == 'kernelpca':
    #     dr = KernelPCA(**content)
    # elif key.lower() == 'isomap':
    #     dr = Isomap(**content)
    # elif key.lower() == 'lle':
    #     dr = LocallyLinearEmbedding(**content)
    # elif key.lower() == 'ltsa':
    #     dr = LocallyLinearEmbedding(**content)
    # elif key.lower() == 'se':
    #     dr = SpectralEmbedding(**content)
    # elif key.lower() == 'mds':
    #     dr = MDS(**content)
    # elif key.lower() == 'tsne':
    #     dr = TSNE(**content)
    # else:
    #     dr = DummyNone(**content)
    # return (key, dr)


def parse_clustering(key, content):
    """Parse the options of the clustering step.

    This function does the same as parse_preproc but works on the clustering
    options.

    Parameters
    -----------
    key : {'KMeans', 'KernelKMeans', 'AP', 'MS', 'Spectral', 'Hierarchical'}
        The selected dimensionality reduction algorithm.

    content : dict
        A dictionary containing parameters for each clustering class.
        Each parameter can be a list; in this case for each combination
        of parameters a different pipeline will be created.

    Returns
    -----------
    cltpl : tuple
        A tuple made like that ('clust_name', clust_obj), where clust_obj
        implements the .fit method.
    """
    if 'auto' in [content.get('n_clusters', ''), content.get('preference', '')]:
        # Wrapper class that automatically detects the best number of clusters
        # via 10-Fold CV
        content.pop('n_clusters', '')
        content.pop('preference', '')

        kwargs = {'param_grid': [], 'n_jobs': -1,
                  'scoring': silhouette_score, 'cv': 10}

        if key.lower() == 'kmeans':
            content.setdefault('init', 'k-means++')
            content.setdefault('n_jobs', 1)
            kwargs['estimator'] = KMeans(**content)
        elif key.lower() == 'ap':
            kwargs['estimator'] = AffinityPropagation(**content)
            kwargs['affinity'] = kwargs['estimator'].affinity
        else:
            logging.warning("n_clusters = 'auto' specified outside kmeans or ap."
                            " Creating GridSearchCV pipeline anyway ...")
        cl = GridSearchCV(**kwargs)

    else:
        if key.lower() == 'kmeans':
            content.setdefault('n_jobs', -1)
            cl = KMeans(**content)
        elif key.lower() == 'ap':
            content.setdefault('preference', 1)
            cl = AffinityPropagation(**content)
        elif key.lower() == 'ms':
            cl = MeanShift(**content)
        elif key.lower() == 'spectral':
            cl = SpectralClustering(**content)
        elif key.lower() == 'hierarchical':
            cl = AgglomerativeClustering(**content)
        else:
            cl = DummyNone()
    return (key, cl)


def parse_steps(steps, max_n_pipes=100):
    """Parse the steps and create the pipelines.

    This function parses the steps coded as dictionaries in the ade_config
    files and creates a sklearn pipeline objects for each combination of
    imputing -> preprocessing -> dimensinality reduction -> clustering
    algorithms.

    A typical step may be of the following form:
        stepX = {'Algorithm': [On/Off flag, [variant0, ...]]}
    where On/Off flag = {True, False} and variantX = 'string'.

    Parameters
    -----------
    steps : list of dictionaries
        A list of (usually 4) dictionaries that contains the details of the
        pipelines to implement.

    max_n_pipes : int, optional, default: 100
        The maximum number of combinations allowed. This avoids a too expensive
        computation.

    Returns
    -----------
    pipes : list of sklearn.pipeline.Pipeline
        The returned list must contain every possible combination of
        imputing -> preprocessing -> dimensionality reduction -> clustering
        algorithms (up to max_n_pipes).
    """
    pipes = []       # a list of list of tuples input of sklearn Pipeline
    imputing, preproc, dimred, clustering = steps[:4]

    # Parse the imputing options
    i_lst_of_tpls = []
    if imputing['Impute'][0]:  # On/Off flag
        if len(imputing['Impute']) > 1:
            content_d = imputing['Impute'][1]
            content_values = values_iterator(content_d)
            for ll in modified_cartesian(*map(ensure_list, list(content_values))):
                content = {__k: __v for __k, __v in zip(list(content_d), ll)}
                i_lst_of_tpls.append(("Impute", Imputer(**content)))
        else:
            i_lst_of_tpls.append(("Impute", Imputer()))

    # Parse the preprocessing options
    pp_lst_of_tpls = []
    for key in preproc.keys():
        if preproc[key][0]:  # On/Off flag
            if len(preproc[key]) > 1:
                content_d = preproc[key][1]
                content_values = values_iterator(content_d)
                for ll in modified_cartesian(*map(ensure_list, list(content_values))):
                    content = {__k: __v for __k, __v in zip(list(content_d), ll)}
                    pp_lst_of_tpls.append(parse_preproc(key, content))
            else:
                pp_lst_of_tpls.append(parse_preproc(key, {}))
                # pp_lst_of_tpls.append(parse_preproc(key, preproc[key]))

    # Parse the dimensionality reduction & manifold learning options
    dr_lst_of_tpls = []
    for key in dimred.keys():
        if dimred[key][0]:  # On/Off flag
            if len(dimred[key]) > 1:
                content_d = dimred[key][1]
                content_values = values_iterator(content_d)
                for ll in modified_cartesian(*map(ensure_list, list(content_values))):
                    content = {__k: __v for __k, __v in zip(list(content_d), ll)}
                    dr_lst_of_tpls.append(parse_dimred(key, content))
            else:
                dr_lst_of_tpls.append(parse_dimred(key, {}))

    # Parse the clustering options
    cl_lst_of_tpls = []
    for key in clustering.keys():
        if clustering[key][0]:  # On/Off flag
            if len(clustering[key]) > 1:  # Discriminate from just flag or flag + args
                content_d = clustering[key][1]
                content_values = values_iterator(content_d)
                for ll in modified_cartesian(*map(ensure_list, list(content_values))):
                    content = {_k: _v for _k, _v in zip(list(content_d), ll)}
                    if not (content.get('affinity', '') in ['manhattan', 'precomputed'] and content.get('linkage', '') == 'ward'):
                        cl_lst_of_tpls.append(parse_clustering(key, content))

            else:  # just flag case
                cl_lst_of_tpls.append(parse_clustering(key, {}))


    #  Generate the list of list of tuples (i.e. the list of pipelines)
    pipes = modified_cartesian(i_lst_of_tpls, pp_lst_of_tpls, dr_lst_of_tpls,
                               cl_lst_of_tpls, pipes_mode=True)
    for pipe in pipes:
        logging.info("Generated pipeline: \n {} \n".format(pipe))
    logging.info("*** {} pipeline(s) generated ***".format(len(pipes)))

    #  Get only the first max_n_pipes
    if len(pipes) > max_n_pipes:
        logging.warning("Maximum number of pipelines reached. "
                        "I'm keeping the first {}".format(max_n_pipes))
        pipes = pipes[0:max_n_pipes]

    return pipes



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""This module is just a wrapper for some sklearn.datasets functions"""

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import numpy as np
import pandas as pd
from sklearn import datasets
from sklearn.preprocessing import Binarizer

# Legacy import
try:
    from sklearn.model_selection import StratifiedShuffleSplit
except ImportError:
    from sklearn.cross_validation import StratifiedShuffleSplit


def generate_gauss(mu=(), std=(), n_sample=()):
    """Create a Gaussian dataset.

    Generates a dataset with n_sample * n_class examples and n_dim dimensions.

    Parameters
    -----------
    mu : array of float, shape : n_class x n_dim
        The mean of each class.

    std :  array of float, shape : n_class
        The standard deviation of each Gaussian distribution.

    n_sample : int
        Number of point per class.
    """
    n_class, n_var = mu.shape

    X = np.zeros((n_sample*n_class, n_var))
    y = np.zeros(n_sample*n_class, dtype=int)

    start = 0
    for i, s, m in zip(range(n_class), std, mu):
        end = start + n_sample
        X[start:end, :] = s * np.random.randn(n_sample, n_var) + m
        y[start:end] = i
        start = end

    return X, y


def load_custom(x_filename, y_filename):
    """Load a custom dataset.

    This function loads the data matrix and the label vector returning a
    unique sklearn-like object dataSetObj.

    Parameters
    -----------
    x_filename : string
        The data matrix file name.

    y_filename : string
        The label vector file name.

    Returns
    -----------
    data : sklearn.datasets.base.Bunch
        An instance of the sklearn.datasets.base.Bunch class, the meaningful
        attributes are .data, the data matrix, and .target, the label vector.
    """
    if x_filename is None:
        raise IOError("Filename for X must be specified with mode 'custom'.")
    if x_filename.endswith('.npy'):  # it an .npy file is provided
        try:  # labels are not mandatory
            y = np.load(y_filename)
        except IOError as e:
            y = np.nan
            e.strerror = "No labels file provided"
            print("I/O error({0}): {1}".format(e.errno, e.strerror))

        return datasets.base.Bunch(data=np.load(x_filename), target=y)
        # return dataSetObj(np.load(x_filename),y)

    elif x_filename.endswith('.csv') or x_filename.endswith('.txt'):
        try:
            dfx = pd.io.parsers.read_csv(x_filename, header=0, index_col=0)
        except IOError as e:
            e.strerror = "Can't open {}".format(x_filename)
            print("I/O error({0}): {1}".format(e.errno, e.strerror))
        y = None
        if y_filename is not None:
            y = pd.io.parsers.read_csv(y_filename,
                                       header=0,
                                       index_col=0).as_matrix().ravel()
        return datasets.base.Bunch(data=dfx.as_matrix(), target=y)


def load(opt='custom', x_filename=None, y_filename=None, n_samples=0):
    """Load a specified dataset.

    This function can be used either to load one of the standard scikit-learn
    datasets or a different dataset saved as X.npy Y.npy in the working
    directory.

    Parameters
    -----------
    opt : {'iris', 'digits', 'diabetes', 'boston', 'circles', 'moons',
          'custom'}, default: 'custom'
        Name of a predefined dataset to be loaded.

    x_filename : string, default : None
        The data matrix file name.

    y_filename : string, default : None
        The label vector file name.

    Returns
    -----------
    X : array of float, shape : n_samples x n_features
        The input data matrix.

    y : array of float, shape : n_samples
        The label vector; np.nan if missing.

    feature_names : array of integers (or strings), shape : n_features
        The feature names; a range of number if missing.
    """
    data = None
    try:
        if opt.lower() == 'iris':
            data = datasets.load_iris()
        elif opt.lower() == 'digits':
            data = datasets.load_digits()
        elif opt.lower() == 'diabetes':
            data = datasets.load_diabetes()
            b = Binarizer(threshold=np.mean(data.target))
            data.target = b.fit_transform(data.data)
        elif opt.lower() == 'boston':
            data = datasets.load_boston()
            b = Binarizer(threshold=np.mean(data.target))
            data.target = b.fit_transform(data.data)
        elif opt.lower() == 'gauss':
            means = np.array([[-1, 1, 1, 1], [0, -1, 0, 0], [1, 1, -1, -1]])
            sigmas = np.array([0.33, 0.33, 0.33])
            if n_samples <= 1: n_samples = 333
            xx, yy = generate_gauss(mu=means, std=sigmas, n_sample=n_samples)
            data = datasets.base.Bunch(data=xx, target=yy)
        elif opt.lower() == 'circles':
            if n_samples == 0: n_samples = 400
            xx, yy = datasets.make_circles(n_samples=n_samples, factor=.3,
                                           noise=.05)
            data = datasets.base.Bunch(data=xx, target=yy)
        elif opt.lower() == 'moons':
            if n_samples == 0: n_samples = 400
            xx, yy = datasets.make_moons(n_samples=n_samples, noise=.01)
            data = datasets.base.Bunch(data=xx, target=yy)
        elif opt.lower() == 'custom':
            data = load_custom(x_filename, y_filename)
    except IOError as e:
        print("I/O error({0}): {1}".format(e.errno, e.strerror))

    X, y = data.data, data.target
    if y is not None and n_samples > 0 and X.shape[0] > n_samples:
        try:  # Legacy for sklearn
            sss = StratifiedShuffleSplit(y, test_size=n_samples, n_iter=1)
            # idx = np.random.permutation(X.shape[0])[:n_samples]
        except TypeError:
            sss = StratifiedShuffleSplit(n_iter=1, test_size=n_samples) \
                  .split(X, y)

        _, idx = list(sss)[0]
        X, y = X[idx, :], y[idx]

    try:
        feat_names = data.features_names
    except:
        feat_names = range(0, X.shape[1])
    try:
        class_names = data.target_names
    except:
        class_names = 0

    return X, y, feat_names, class_names



#!/usr/bin/python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import os
import copy
import logging
import multiprocessing
import cPickle as pkl
import numpy as np
# from adenine.utils.extra import make_time_flag
from adenine.utils.extra import get_time
from adenine.utils.extra import timed


def create(pdef):
    """Scikit-learn Pipelines objects creation (deprecated).

    This function creates a list of sklearn Pipeline objects starting from the
    list of list of tuples given in input that could be created using the
    adenine.core.define_pipeline module.

    Parameters
    -----------
    pdef : list of list of tuples
        This arguments contains the specification needed by sklearn in order
        to create a working Pipeline object.

    Returns
    -----------
    pipes : list of sklearn.pipeline.Pipeline objects
        The list of Piplines, each of them can be fitted and trasformed
        with some data.
    """
    from sklearn.pipeline import Pipeline
    pipes = []
    for p in pdef:
        pipes.append(Pipeline(p))
    return pipes


def which_level(label):
    """Define the step level according to the input step label.

    This function return the level (i.e.: imputing, preproc, dimred, clustring,
    None) according to the step label provided as input.

    Parameters
    -----------
    label : string
        This is the step level as it is reported in the ade_config file.

    Returns
    -----------
    level : {imputing, preproc, dimred, clustering, None}
        The appropriate level of the input step.
    """
    if label in ('Impute_median', 'Impute_mean', 'Impute'):
        level = 'imputing'
    elif label in ('Recenter', 'Standardize', 'Normalize', 'MinMax'):
        level = 'preproc'
    elif label in ('PCA', 'IncrementalPCA', 'RandomizedPCA', 'KernelPCA',
                   'Isomap', 'LLE', 'SE', 'MDS', 'tSNE'):
        level = 'dimred'
    elif label in ('KMeans', 'KernelKMeans', 'AP', 'MS', 'Spectral',
                   'Hierarchical'):
        level = 'clustering'
    else:
        level = 'None'
    return level


def evaluate(level, step, X):
    """Transform or predict according to the input level.

    This function uses the transform or the predict method on the input
    sklearn-like step according to its level (i.e. imputing, preproc, dimred,
    clustering, none).

    Parameters
    -----------
    level : {'imputing', 'preproc', 'dimred', 'clustering', 'None'}
        The step level.

    step : sklearn-like object
        This might be an Imputer, or a PCA, or a KMeans (and so on...)
        sklearn-like object.

    X : array of float, shape : n_samples x n_features
        The input data matrix.

    Returns
    -----------
    res : array of float
        A matrix projection in case of dimred, a label vector in case of
        clustering, and so on.
    """
    if level in ['imputing', 'preproc', 'dimred', 'None']:
        if hasattr(step, 'embedding_'):
            res = step.embedding_
        else:
            res = step.transform(X)
    elif level == 'clustering':
        if hasattr(step, 'labels_'):
            res = step.labels_  # e.g. in case of spectral clustering
        else:
            res = step.predict(X)
    return res


def pipe_worker(pipeID, pipe, pipes_dump, X):
    """Parallel pipelines execution.

    Parameters
    -----------
    pipeID : string
        Pipeline identifier.

    pipe : list of tuples
        Tuple containing a label and a sklearn Pipeline object.

    pipes_dump : multiprocessing.Manager.dict
        Dictionary containing the results of the parallel execution.

    X : array of float, shape : n_samples x n_features, default : ()
        The input data matrix.
    """
    step_dump = dict()

    # COPY X as X_curr (to avoid that the next pipeline
    # works on the results of the previuos one)
    X_curr = np.array(X)
    for j, step in enumerate(pipe):
        # step[0] -> step_label | step[1] -> model, sklearn (or sklearn-like)
        # object
        stepID = 'step'+str(j)
        # 1. define which level of step is this (i.e.: imputing, preproc,
        # dimred, clustering, none)
        level = which_level(step[0])
        # 2. fit the model (whatever it is)
        if step[1].get_params().get('method') == 'hessian':
            # check hessian lle constraints
            n_components = step[1].get_params().get('n_components')
            step[1].set_params(n_neighbors=1+(n_components*(n_components+3)/2))
        try:
            step[1].fit(X_curr)

            # 3. evaluate (i.e. transform or predict according to the level)
            # X_curr = evaluate(level, step[1], X_curr)
            X_next = evaluate(level, step[1], X_curr)
            # 3.1 if the model is suitable for voronoi tessellation: fit also
            # on 2D
            mdl_voronoi = None
            if hasattr(step[1], 'cluster_centers_'):
                if hasattr(step[1], 'best_estimator_'):
                    mdl_voronoi = copy.copy(step[1].best_estimator_)
                else:
                    mdl_voronoi = copy.copy(step[1])
                mdl_voronoi.fit(X_curr[:, :2])

            # 4. save the results in a dictionary of dictionaries of the form:
            # {'pipeID': {'stepID' : [alg_name, level, params, res, Xnext, Xcurr, stepObj, voronoi_suitable_model]}}
            if level in ('preproc', 'imputing'):  # save memory and do not dump data after preprocessing (not used in analysys)
                step_dump[stepID] = [step[0], level, step[1].get_params(), np.array([]), np.array([]), step[1], mdl_voronoi]
                X_curr = np.array(X_next)  # update the matrix
            elif level == 'dimred':  # save memory dumping X_curr only in case of clusutering
                step_dump[stepID] = [step[0], level, step[1].get_params(), X_next, np.array([]), step[1], mdl_voronoi]
                X_curr = X_next  # update the matrix
            elif level == 'clustering':  # clustering
                step_dump[stepID] = [step[0], level, step[1].get_params(), X_next, X_curr, step[1], mdl_voronoi]

            # if level != 'clustering':
            #     X_curr = np.array(X_next) # update the matrix

        except (AssertionError, ValueError) as e:
            logging.critical("Pipeline {} failed at step {}. "
                             "Traceback: {}".format(pipeID, step[0], e))
            step_dump[stepID] = [step[0], level, step[1].get_params(), np.nan,
                                 np.nan, np.nan]

    pipes_dump[pipeID] = step_dump
    logging.debug("DUMP: \n {} \n #########".format(pipes_dump))


@timed
def run(pipes=(), X=(), exp_tag='def_tag', root='', y=None):
    """Fit and transform/predict some pipelines on some data.

    This function fits each pipeline in the input list on the provided data.
    The results are dumped into a pkl file as a dictionary of dictionaries of
    the form {'pipeID': {'stepID' : [alg_name, level, params, data_out,
    data_in, model_obj, voronoi_suitable_object], ...}, ...}. The model_obj is
    the sklearn model which has been fit on the dataset, the
    voronoi_suitable_object is the very same model but fitted on just the first
    two dimensions of the dataset. If a pipeline fails for some reasons the
    content of the stepID key is a list of np.nan.

    Parameters
    -----------
    pipes : list of list of tuples
        Each tuple contains a label and a sklearn Pipeline object.

    X : array of float, shape : n_samples x n_features, default : ()
        The input data matrix.

    exp_tag : string
        An intuitive tag for the current experiment.

    root : string
        The root folder to save the results.

    y : array-like, optional
        If specified, it contains data labels.

    Returns
    -----------
    outputFolderName : string
        The path of the output folder.
    """
    # Check root folder
    if not os.path.exists(root):  # if it does not exist
        if not len(root):         # (and the name has not been even specified)
            root = 'results_'+exp_tag+get_time()  # create a standard name
        os.makedirs(root)         # and make the folder
        logging.warn("No root folder supplied, "
                     "folder {} created".format(os.path.abspath(root)))

    jobs = []
    manager = multiprocessing.Manager()
    pipes_dump = manager.dict()

    # Submit jobs
    for i, pipe in enumerate(pipes):
        pipeID = 'pipe'+str(i)
        p = multiprocessing.Process(target=pipe_worker,
                                    args=(pipeID, pipe, pipes_dump, X))
        jobs.append(p)
        p.start()
        logging.info("Job: {} submitted".format(pipeID))

    # Collect results
    ret_count = 0
    for proc in jobs:
        proc.join()
        ret_count += 1
    logging.info("{} jobs collected".format(ret_count))

    # Convert the DictProxy to standard dict
    pipes_dump = dict(pipes_dump)

    # Output Name
    output_filename = exp_tag
    output_folder = os.path.join(root, output_filename)

    # Create exp folder into the root folder
    os.makedirs(output_folder)

    # pkl Dump
    import gzip
    with gzip.open(os.path.join(output_folder,
                                output_filename+'.pkl.tz'), 'w+') as f:
        pkl.dump(pipes_dump, f)
    logging.info("Dumped : {}".format(os.path.join(output_folder,
                                                   output_filename+'.pkl.tz')))
    with gzip.open(os.path.join(output_folder, '__data.pkl.tz'), 'w+') as f:
        pkl.dump({'X': X, 'y': y}, f)
    logging.info("Dumped : {}".format(os.path.join(output_folder,
                                                   '__data.pkl.tz')))

    return output_folder



#!/usr/bin/python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import logging
import warnings

import numpy as np

from sklearn.preprocessing import Imputer
from sklearn.neighbors import NearestNeighbors
from sklearn.metrics.pairwise import pairwise_distances

from sklearn.metrics import silhouette_score as sil

# Legacy import
try:
    from sklearn.model_selection import GridSearchCV
except ImportError:
    from sklearn.grid_search import GridSearchCV


class DummyNone:
    """Dummy class that does nothing.

    It is a sklearn 'transforms', it implements both a fit and a transform
    method and it just returns the data in input. It has been created only for
    consistency with sklearn.
    """

    def __init__(self, **kwargs):
        if kwargs.get('n_components', 0) > 0:
            self.n_components = kwargs.get('n_components')

    def fit(self, X):
        return self

    def transform(self, X):
        return X

    def get_params(self):
        return dict()


class Imputer(Imputer):
    """Extension of the sklearn.preprocessing.Imputer class.

    This class adds the nearest_neighbors data imputing strategy.
    """

    def fit(self, X, y=None):
        if self.strategy.lower() in ['nearest_neighbors', 'nn']:
            self._nn_fit(X)
        else:
            if y:
                super(Imputer, self).fit(X, y)
            else:
                super(Imputer, self).fit(X)
        return self

    def transform(self, X):
        if self.strategy.lower() in ['nearest_neighbors', 'nn']:
            _X = X[self._mask, :].copy()
            _X[self.missing] = self.statistics_[self.missing]
            return _X
        else:
            return super(Imputer, self).transform(X)

    def _get_mask(self, X, value_to_mask):
        """Compute the boolean mask X == missing_values.
        [copy/pasted from sklearn.preprocessing]"""
        if value_to_mask == "NaN" or np.isnan(value_to_mask):
            return np.isnan(X)
        else:
            return X == value_to_mask

    def _get_row_indexes(self, i, c_idx):
        """
        Get which samples do not have missing values or have the same missing
        value as the i-th sample.
        """
        # Drop the column with missing values in the i-th sample
        _missing = self.missing.copy()
        _missing = _missing[:, c_idx]

        # Get the filled columns
        r_idx = []
        for k, r in enumerate(_missing):
            _not_row = [not j for j in r]
            if np.prod(_not_row, dtype=np.bool):  # it's like False is not in _not_row
                r_idx.append(k)

        return np.array(r_idx)

    def _filling_worker(self, X, row, i):
        """
        Worker for parallel execution of self._nn_fit()
        """

        # the list of non-missing values for the i-th row
        c_idx = np.where([not j for j in row])[0]

        # Generate the training matrix (only the non-empty columns)
        r_idx = self._get_row_indexes(i, c_idx)
        Xtr = X[r_idx[:, np.newaxis], c_idx]  # get the full matrix of possible neighbors

        neigh = NearestNeighbors(n_neighbors=min(6, Xtr.shape[0]), n_jobs=1)

        # Get the nearest Neighbors
        neigh.fit(Xtr)

        with warnings.catch_warnings():  # shut-up deprecation warnings
            warnings.simplefilter("ignore")
            _nn_idx = neigh.kneighbors(X[i, c_idx], return_distance=False)

        _nn_idx = _nn_idx[0]

        # Evaluate the average of the nearest Neighbors
        neighbors = X[r_idx[_nn_idx[1:]], :]  # matrix of nearest neighbors, skip the first one which is the same

        with warnings.catch_warnings():  # shut-up deprecation warnings
            warnings.simplefilter("ignore")
            _nanmean = np.nanmean(neighbors[:, np.where(row)[0]], axis=0)

        self.statistics_[i, row] = _nanmean

    def _nn_fit(self, X):
        """Impute the input data matrix using the Nearest Neighbors approach.

        This implementation follows, approximately, the strategy proposed in:
        [Hastie, Trevor, et al. "Imputing missing data for gene expression arrays." (1999): 1-7.]
        [Troyanskaya, Olga, et al. "Missing value estimation methods for DNA microarrays." Bioinformatics 17.6 (2001): 520-525.]
        """
        # 1. Find missing values

        self.missing = self._get_mask(X, self.missing_values)
        # 2. Drop empty rows (I cannot deal with that)
        _mask = np.prod(self.missing, axis=1, dtype=np.bool)
        self._mask = np.array([not j for j in _mask])
        _X = X[self._mask,:].copy()
        self.missing = self.missing[self._mask,:]

        # 3. Statistics init
        self.statistics_ = np.empty_like(_X)

        # 4. For each row that presents a True value in missing:
        #       drop the True column and get the first K Nearest Neighbors
        _cond = True
        count = 0
        while _cond and count < 100:
            for i, row in enumerate(self.missing):
                if row.any(): # i.e. if True in row:
                    self._filling_worker(_X, row, i)
            _cond = np.isnan(self.statistics_).any()
            _X[self.missing] = self.statistics_[self.missing]
            count += 1

        # Log the failure
        if _cond:
            logging.info("Data imputing partially failed.")

        return self


class GridSearchCV(GridSearchCV):
    """
    Wrapper class that automatically detects the optimal number of clusters for centroid based algorithm like KMeans and Affinity Propagation.
    """
    def __init__(self, estimator, param_grid, scoring=None, fit_params=None, n_jobs=1, iid=True, refit=True, cv=None, verbose=0, pre_dispatch='2*n_jobs', error_score='raise', affinity='euclidean'):
        super(GridSearchCV, self).__init__( estimator, param_grid, scoring, fit_params, n_jobs, iid, refit, cv, verbose, pre_dispatch, error_score)
        self.affinity = affinity # add the attribute affinity
        self.cluster_centers_ = None
        self.inertia_ = None
        self.n_clusters = None
        self.estimator_name = type(self.estimator).__name__

    def _sqrtn_heuristic(self, n):
        """
        n_clusters grid for KMeans: logaritmic scale in [2,...,log10(sqrt(n))] with max length = 30
        """
        return np.unique(map(int, np.logspace(np.log10(2), np.log10(np.sqrt(n)) ,30)))

    def _min_max_dist_heuristic(self, X, affinity):
        """
        preference grid for Affinity Propagation: linear scale in [min(similarity matrix),...,median(similarity matrix)]
        """
        S = -pairwise_distances(X, metric=affinity, squared=True)
        return np.unique(map(int, np.linspace(np.min(S), np.median(S), 30)))

    def fit(self, X, y=None):
        """
        This new definition of the fit method sets the grid follwing a different heuristic according to the clustering algorithm
        """
        # Pick the heuristic
        if type(self.estimator).__name__ == 'KMeans':
            # pick heuristic 1
            self.param_grid = {'n_clusters': self._sqrtn_heuristic(X.shape[0])}
        elif type(self.estimator).__name__ == 'AffinityPropagation':
            # pick heuristic 2
            self.param_grid = {'preference': self._min_max_dist_heuristic(X, self.affinity)}

        # Then perform standard fit
        if y:
            super(GridSearchCV, self).fit(X,y)
        else:
            super(GridSearchCV, self).fit(X)

        # Propagate the cluster_centers_ attribute (needed for voronoi plot)
        if hasattr(self.best_estimator_, 'cluster_centers_'): # added for consistency only
            self.cluster_centers_ = self.best_estimator_.cluster_centers_

        # Propagate the inertia_ attribute
        if hasattr(self.best_estimator_, 'inertia_'):
            self.inertia_ = self.best_estimator_.inertia_

        # Propagate the n_clusters
        if hasattr(self.best_estimator_, 'cluster_centers_'):
            self.n_clusters = self.best_estimator_.cluster_centers_.shape[0]

        if hasattr(self.best_estimator_, 'affinity_matrix_'):
            self.affinity_matrix_ = self.best_estimator_.affinity_matrix_

        return self

    def get_params(self, deep=True):
        params_ = super(GridSearchCV, self).get_params(deep)
        params_['n_clusters'] = self.n_clusters
        params_['inertia_'] = self.inertia_
        params_['estimator_name'] = self.estimator_name
        return params_



def silhouette_score(estimator, X, y=None):
    """
    scorer wrapper for metrics.silhouette_score
    """
    _y = estimator.predict(X)
    n_labels = len(np.unique(_y))
    if 1 < n_labels < X.shape[0]:
        return sil(X, _y)
    else:
        logging.info("adenine.utils.extension.silhouette_score() returned NaN because the number of labels is {}. Valid values are 2 to n_samples - 1 (inclusive) = {}".format(n_labels, X.shape[0]-1))
        return np.nan



#!/usr/bin/python
# adenine setup script

from distutils.core import setup

# Package Version
from adenine import __version__ as version

setup(
    name='adenine',
    version=version,

    description=('A Data ExploratioN pIpeliNE'),
    long_description=open('README.md').read(),
    author='Samuele Fiorini, Federico Tomasi',
    author_email='{samuele.fiorini, federico.tomasi}@dibris.unige.it',
    maintainer='Samuele Fiorini, Federico Tomasi',
    maintainer_email='{samuele.fiorini, federico.tomasi}@dibris.unige.it',
    url='https://github.com/slipguru/adenine',
    download_url = 'https://github.com/slipguru/adenine/tarball/'+version,
    classifiers=[
	'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'License :: OSI Approved :: BSD License',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS'
    ],
    license = 'FreeBSD',

    packages=['adenine', 'adenine.core', 'adenine.utils'],
    requires=['numpy (>=1.10.1)',
              'scipy (>=0.16.1)',
              'sklearn (>=0.17)',
              'matplotlib (>=1.5.1)',
              'seaborn (>=0.7.0)',
              'fastcluster (>=1.1.20)'],
    scripts=['scripts/ade_run.py','scripts/ade_analysis.py'],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from adenine.utils import data_source
from adenine.utils import extra

# --------------------------  EXPERMIENT INFO ------------------------- #
exp_tag = 'cool_experiment'
output_root_folder = 'results'
plotting_context = 'notebook' # one of {paper, notebook, talk, poster}
file_format = 'pdf' # or 'png'

# ----------------------------  INPUT DATA ---------------------------- #
data_file = 'data.csv'
labels_file = 'labels.csv' # OPTIONAL
X, y, feat_names, class_names = data_source.load('custom', data_file, labels_file)

# -----------------------  PIPELINE DEFINITION ------------------------ #

# --- Missing Values Imputing --- #
step0 = {'Impute': [False, {'missing_values': 'NaN',
                            'strategy': ['median','mean','nearest_neighbors']}]}

# --- Data Preprocessing --- #
step1 = {'None': [False], 'Recenter': [False], 'Standardize': [False],
         'Normalize': [False, {'norm': ['l1','l2']}],
         'MinMax': [False, {'feature_range': [(0,1), (-1,1)]}]}

# --- Dimensionality Reduction & Manifold Learning --- #
step2 = {'PCA': [False, {'n_components': 3}],
         'IncrementalPCA': [False],
         'RandomizedPCA': [False],
         'KernelPCA': [False, {'kernel': ['linear','rbf','poly']}],
         'Isomap': [False, {'n_neighbors': 5}],
         'LLE': [False, {'n_neighbors': 5, 'method': ['standard','modified','hessian','ltsa']}],
         'SE': [False, {'affinity': ['nearest_neighbors','rbf']}],  # affinity can be 'precomputed'
         'MDS': [False, {'metric': True}],
         'tSNE': [False],
         'None': [False]
         }

# --- Clustering --- #
step3 = {'KMeans': [False, {'n_clusters': [3, 'auto']}],
         'AP': [False, {'preference': ['auto']}],                        # affinity can be 'precomputed'
         'MS': [False],                                                  # affinity can be 'precomputed'
         'Spectral': [False, {'n_clusters': [3, 8]}],                    # affinity can be 'precomputed'
         'Hierarchical': [False, {'n_clusters': [3, 8],
                                  'affinity': ['manhattan','euclidean'], # affinity can be 'precomputed'
                                  'linkage':  ['ward','complete','average']}]
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-

######################################################################
# Copyright (C) 2016 Samuele Fiorini, Federico Tomasi, Annalisa Barla
#
# FreeBSD License
######################################################################

import imp, sys, os, shutil
import time
import logging
import argparse

from adenine.core import define_pipeline
from adenine.core import pipelines
from adenine.utils import extra

def main(config_file):
    # Load the configuration file
    config_path = os.path.abspath(config_file)
    config = imp.load_source('ade_config', config_path)
    extra.set_module_defaults(config, {'step0': {'Impute': [False]},
                                       'step1': {'None': [True]},
                                       'step2': {'None': [True]},
                                       'step3': {'None': [False]},
                                       'exp_tag': 'debug',
                                       'output_root_folder': 'results'})

    # Read the variables from the config file
    X, y, feat_names, class_names = config.X, config.y, config.feat_names, config.class_names

    # Get the experiment tag and the output root folder
    exp_tag, root = config.exp_tag, config.output_root_folder
    if not os.path.exists(root):
        os.makedirs(root)

    # Define the ade.log file (a new one for each run)
    fileName = '_'.join(('ade', exp_tag, extra.get_time()))
    logFileName = os.path.join(root, fileName+'.log')
    logging.basicConfig(filename=logFileName, level=logging.INFO, filemode='w',
                        format='%(levelname)s (%(name)s): %(message)s')
    root_logger = logging.getLogger()
    ch = logging.StreamHandler()
    ch.setLevel(logging.CRITICAL)
    ch.setFormatter(logging.Formatter('%(levelname)s (%(name)s): %(message)s'))
    root_logger.addHandler(ch)

    # Pipelines Definition
    pipes = define_pipeline.parse_steps([config.step0, config.step1,
                                             config.step2, config.step3])

    # Pipelines Evaluation
    outFolder = pipelines.run(pipes=pipes, X=X, exp_tag=fileName, root=root, y=y)

    # Copy the ade_config just used into the outFolder
    shutil.copy(config_path, os.path.join(outFolder, 'ade_config.py'))
    # Move the logging file into the outFolder
    shutil.move(logFileName, outFolder)

# ----------------------------  RUN MAIN ---------------------------- #
if __name__ == '__main__':
    from adenine import __version__
    parser = argparse.ArgumentParser(#usage="%(prog)s [-c] configuration_file.py",
                                     description='Adenine script for generating pipelines.')
    parser.add_argument('--version', action='version', version='%(prog)s v'+__version__)
    parser.add_argument("-c", "--create", dest="create", action="store_true",
                        help="create config file", default=False)
    parser.add_argument("configuration_file", help="specify config file", default='ade_config.py')
    args = parser.parse_args()

    if args.create:
        import adenine as ade
        std_config_path = os.path.join(ade.__path__[0], 'ade_config.py')
        # Check for .pyc
        if std_config_path.endswith('.pyc'):
            std_config_path = std_config_path[:-1]
        # Check if the file already exists
        if os.path.exists(args.configuration_file):
            parser.error("adenine configuration file already exists")
        # Copy the config file
        shutil.copy(std_config_path, args.configuration_file)
    else:
        main(args.configuration_file)
